import inspect
from typing import Union, get_origin, get_args
import yaml
from abc import ABC


class YamlAttributes(ABC):
    yaml_file_path: str = './yaml-attribute-config.yaml'
    yaml_section: str = 'config'

    @classmethod
    def init(cls, mode='sync', yaml_file_path=None, yaml_section=None):
        cls.__load_config(
            yaml_file_path or cls.yaml_file_path,
            yaml_section or cls.yaml_section,
            mode,
        )

    @classmethod
    def to_dict(cls):
        return {
            k: v
            for k, v in vars(cls).items()
            if not (
                k.startswith('_')
                or inspect.ismethod(k)
            )
        }

    @classmethod
    def __load_config(cls, yam_file_path: str, yaml_section, mode):
        attributes = cls.__get_attributes()

        with open(yam_file_path, "r") as stream:
            config = yaml.safe_load(stream)

            if yaml_section:
                config = config[yaml_section]

            cls.__set_attributes(config, attributes, mode)

    @classmethod
    def __get_attributes(cls):
        self_members = dict(inspect.getmembers(YamlAttributes))
        members = dict(inspect.getmembers(cls))

        members = {
            **members,
            **members['__annotations__'],
        }

        filtered_members = {
            key: value
            for (key, value) in members.items()
            if not (
                key.startswith('_')
                or inspect.ismethod(value)
                or key in self_members.keys()
            )
        }

        return filtered_members

    @classmethod
    def __set_attributes(cls, config: dict, attributes: dict, mode):
        req_attributes = [
            k
            for k, v in attributes.items()
            if not cls.__is_optional(v)
        ]

        modes = {
            # The YAML config and the required class attributes have to match
            # exactly while the YAML config can only have entries which are
            # also in the config class
            'sync': lambda: (
                all(k in config for k in req_attributes)
                and all(k in attributes for k in config.keys())
            ),
            # The YAML config has to have at least all class attributes
            # while additional entries are omitted
            'soft_config': lambda: all(k in config for k in req_attributes)
        }

        assert (
            modes[mode]()
        ), 'YAML config and/or class attributes do not fulfill the '\
            'requirements of the "{}" mode'.format(mode)

        for key, attribute_type in attributes.items():
            value = config[key] if key in config.keys() else None
            config_type = type(value)
            attribute_is_union = get_origin(attribute_type) is Union

            if not (
                isinstance(attribute_type, type)
                or attribute_is_union
            ):
                attribute_type = type(attributes[key])

            assert (
                config_type == attribute_type
                or (
                    attribute_is_union
                    and config_type in get_args(attributes[key])
                )
            ), 'Type missmatch between YAML file and config '\
                'class was found for the "{}" attribute.'.format(key)

            setattr(cls, key, value)

    def __is_optional(field):
        return get_origin(field) is Union and type(None) in get_args(field)



from typing import Optional
import pytest

from yamlattributes import YamlAttributes


def test_config_loads_successfully_in_sync_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act
    TestConfig.init()

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
    ), 'Config values are not correctly assigned to config class'


def test_config_fails_with_incomplete_yaml_in_sync_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/incomplete_test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act

    # Assert
    with pytest.raises(
        AssertionError,
        match=r".*\"sync\" mode.*",
    ):
        TestConfig.init()


def test_config_fails_with_overloaded_yaml_in_sync_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/overloaded_test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act

    # Assert
    with pytest.raises(
        AssertionError,
        match=r".*\"sync\" mode.*",
    ):
        TestConfig.init()


def test_config_loads_successfully_in_soft_config_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act
    TestConfig.init(mode='soft_config')

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
    ), 'Config values are not correctly assigned to config class'


def test_config_loads_successfully_overloaded_config_in_soft_config_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/overloaded_test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act
    TestConfig.init(mode='soft_config')

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
        and not hasattr(TestConfig, 'test_attribute_d')
    ), 'Config values are not correctly assigned to config class'


def test_config_fails_with_incomplete_yaml_in_soft_config_mode():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/incomplete_test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act

    # Assert
    with pytest.raises(
        AssertionError,
        match=r".*\"soft_config\" mode.*",
    ):
        TestConfig.init(mode='soft_config')


def test_config_fails_with_type_mismatch():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        # create type mismatch between config class and YAML file
        # by changing str to int
        test_attribute_a: int
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act

    # Assert
    with pytest.raises(
        AssertionError,
        match=r".*missmatch.*\"test_attribute_a\" attribute.*",
    ):
        TestConfig.init()


def test_config_loads_successfully_with_union_type():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int
        optional_attribute: Optional[int]

    # Act
    TestConfig.init()

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
        and hasattr(TestConfig, 'optional_attribute')
        and not TestConfig.optional_attribute
    ), 'Config values are not correctly assigned to config class'


def test_config_successfully_assighns_optional_values():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: Optional[int]

    # Act
    TestConfig.init()

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
    ), 'Config values are not correctly assigned to config class'


def test_config_successfully_converts_to_dict():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    test_config_dict = {
        'yaml_file_path': './tests/test_config.yaml',
        'test_attribute_a': 'config_test_value_a',
        'test_attribute_b': 'config_test_value_b',
        'test_attribute_c': 42,
    }

    # print(TestConfig.to_dict())

    # Act
    TestConfig.init()

    # Assert
    assert (
        test_config_dict == TestConfig.to_dict()
    ), 'Config class was not correctly converted to a dict'


def test_config_loads_successfully_when_config_path_is_passed_to_init():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act
    TestConfig.init(yaml_file_path='./tests/test_config.yaml')

    # Assert
    assert (
        TestConfig.test_attribute_a == 'config_test_value_a'
        and TestConfig.test_attribute_b == 'config_test_value_b'
        and TestConfig.test_attribute_c == 42
    ), 'Config values are not correctly assigned to config class'


def test_config_loads_successfully_when_config_section_is_passed_to_init():
    YamlAttributes.__abstractmethods__ = set()

    # Arrange
    class TestConfig(YamlAttributes):
        yaml_file_path = './tests/test_config.yaml'
        test_attribute_a: str
        test_attribute_b = 'test_value_b'
        test_attribute_c: int

    # Act
    TestConfig.init(yaml_section='another_config')

    # Assert
    assert (
        TestConfig.test_attribute_a == 'another_config_test_value_a'
        and TestConfig.test_attribute_b == 'another_config_test_value_b'
        and TestConfig.test_attribute_c == 7
    ), 'Config values are not correctly assigned to config class'
